# vim: tabstop=4 shiftwidth=4 softtabstop=4

# Copyright 2017 Lenovo
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

#This handles port forwarding for web interfaces on management devices
#It will also hijack port 3900 and do best effort..

import asyncio
import socket
import time

forwardersbyclient = {}
relaysbysession = {}
sessionsbyip = {}
ipsbysession = {}
sockhandler = {}

webforwarders = {}


async def relay_streams(inlet, outlet):
    currchunk = await inlet.read(1024)
    while currchunk:
        outlet.write(currchunk)
        currchunk = await inlet.read(1024)
    outlet.close()

class WebForwarder(object):
    def __init__(self, target, vrf):
        self.clientips = set([])
        self.target = target
        self.sessionsbyclient = {}
        self.sock = None
        self.vrf = vrf
        if self.vrf and not isinstance(self.vrf, bytes):
            self.vrf = self.vrf.encode('utf8')
        self.clientvintages = {}


    def allow_client(self, clientip, sessid=None):
        if '.' in clientip and not clientip.startswith('::ffff:'):
            clientip = '::ffff:' + clientip
        if sessid:
            if clientip not in self.sessionsbyclient:
                self.sessionsbyclient[clientip] = set([])
            self.sessionsbyclient[clientip].add(sessid)
        self.clientips.add(clientip)
        # expire in one minute from most recent authenticated access
        self.clientvintages[clientip] = time.monotonic() + 60
    
    def close_client(self, clientip, sessid):
        if '.' in clientip and not clientip.startswith('::ffff:'):
            clientip = '::ffff:' + clientip
        if clientip in self.sessionsbyclient:
            if sessid in self.sessionsbyclient[clientip]:
                self.sessionsbyclient[clientip].discard(sessid)
            if not self.sessionsbyclient[clientip]:
                if clientip in self.clientips:
                    self.clientips.discard(clientip)

    def clean_clients(self):
        for client in list(self.clientvintages):
            if self.clientvintages[client] < time.monotonic():
                self.clientips.discard(client)
                del self.clientvintages[client]

    async def start(self):
        self.sock = socket.socket(socket.AF_INET6)
        try:
            self.sock.setsockopt(socket.SOL_SOCKET, 25, b'mgmt')
        except Exception:
            pass
        self.sock.bind(('::', 0, 0, 0))
        return await asyncio.start_server(self.forward_port, sock=self.sock)

    async def forward_port(self, reader, writer):
            clis = reader._transport.get_extra_info('socket')
            clip = clis.getpeername()[0]
            if clip not in self.clientips:
                writer.close()
                return
            targsock = socket.socket(socket.AF_INET6)
            targsock.setblocking(False)
            if self.vrf:
                targsock.setsockopt(socket.SOL_SOCKET, 25, self.vrf)
            ip, scope = self.target.split('%')
            targaddr = (ip, 443, 0, int(scope))
            # using socket.create_connection would be blocking
            # however using the normal open_connection or even sock_connect
            # breaks because they discard half of the addr
            # so bypass the public interfaces and go for the
            # 'internal' _sock_connect for now
            fut = asyncio.get_running_loop().create_future()
            asyncio.get_running_loop()._sock_connect(fut, targsock, targaddr)
            await fut
            tread, twrite = await asyncio.open_connection(sock=targsock)
            asyncio.create_task(relay_streams(reader, twrite))
            asyncio.create_task(relay_streams(tread, writer))


def close_session(sessionid):
    for addr in forwardersbyclient.get(sessionid, []):
        killsock = forwardersbyclient[sessionid][addr]
        sockhandler[killsock].close()
        del sockhandler[killsock]
        killsock.close()
    if sessionid in forwardersbyclient:
        del forwardersbyclient[sessionid]
    for clip in ipsbysession.get(sessionid, ()):
        sessionsbyip[clip].discard(sessionid)
    if sessionid in ipsbysession:
        del ipsbysession[sessionid]
    for relay in list(relaysbysession.get(sessionid, ())):
        conn = relaysbysession[sessionid][relay]
        relay.kill()
        conn.close()
    if sessionid in relaysbysession:
        del relaysbysession[sessionid]

forward_reaper = None

async def reap_forwards():
    while True:
        await asyncio.sleep(15)
        for addr in webforwarders:
            webforwarders[addr].clean_clients()

async def get_port(addr, clientip, vrf):
    global forward_reaper
    if not forward_reaper:
        forward_reaper = asyncio.create_task(reap_forwards())
    if addr in webforwarders:
        pf = webforwarders[addr]
    else:
        pf = WebForwarder(addr, vrf)
        await pf.start()
        webforwarders[addr] = pf
    pf.allow_client(clientip)
    return pf.sock.getsockname()[1], pf