import asyncio
import os
import socket
import struct

RTM_GETNEIGH = 30
RTM_NEWNEIGH = 28
NLM_F_DUMP = 0x300
NLM_F_REQUEST = 1
NLMSG_DONE = 3
AF_BRIDGE = 7
RTN_UNSPEC = 0
RTN_UNICAST = 1
NUD_PERMAMENT = 0x80
RTM_GETLINK = 18
RTM_NEWLINK = 16
IFF_MASTER = 1 << 10
ARPHRD_ETHER = 1
IFLA_IFNAME = 3

def msg_align(length):
    return (length + 3) & ~3


nlhdrsize = struct.calcsize('IHHII')
ndmsgsize = struct.calcsize('BiHBB')
ilen = struct.calcsize('BHiII')
glsize = nlhdrsize + ilen
glmsg = struct.pack('BHiII', socket.AF_PACKET, 0, 0, 0, 0)
# nlmsghdr struct
nlhdr = struct.pack('IHHII', nlhdrsize + ndmsgsize, RTM_GETNEIGH, NLM_F_DUMP|NLM_F_REQUEST, 0, 0)
nlhdr_gl = struct.pack('IHHII', glsize, RTM_GETLINK, NLM_F_REQUEST|NLM_F_DUMP, 0, 0)
ndmsg = struct.pack('BiHBB', 0, 0, 0, 0, 0)
fdbmg = struct.pack('BiHBB', AF_BRIDGE, 0, 0, 0, 0)
neightable = {}
fdb = {}
iptoport = {}
neighvintage = 0
async def refresh_neighbor_table():
    global neightable
    global fdb
    global iptoport
    global neighvintage
    loop = asyncio.get_running_loop()
    s = socket.socket(socket.AF_NETLINK, socket.SOCK_RAW, socket.NETLINK_ROUTE)
    s.bind((0, 0))
    neightable = {}
    fdb = {}
    iptoport = {}
    neighvintage = os.times()[4]
    try:
        pending = True
        await loop.sock_sendall(s, nlhdr + ndmsg)
        while pending:
            pdata = await loop.sock_recv(s, 65536)
            v = memoryview(pdata)
            length = 1
            while length and len(v):
                length, typ = struct.unpack('IH', v[:struct.calcsize('IH')])
                if typ == NLMSG_DONE:
                    pending = False
                    break
                if typ == RTM_NEWNEIGH:
                    ent = v[nlhdrsize:]
                    fam, idx, state, flags, typ = struct.unpack('BiHBB', ent[:ndmsgsize])
                    if typ == RTN_UNICAST:
                        curraddr = None
                        currip = None
                        rta = ent[ndmsgsize:]
                        rtalen = 1
                        while rtalen and len(rta):
                            rtalen, rtatyp = struct.unpack('HH', rta[:4])
                            if rtatyp == 2:
                                curraddr = rta[4:rtalen].tobytes()
                                if len(curraddr) == 20:
                                    curraddr = curraddr[12:]
                            elif rtatyp == 1:
                                currip = rta[4:rtalen].tobytes()
                            rta = rta[msg_align(rtalen):]
                        if curraddr and currip:
                            neightable[currip] = curraddr
                v = v[msg_align(length):]
        idxtoname = {}
        await loop.sock_sendall(s, nlhdr_gl + glmsg)
        pending = True
        while pending:
            pdata = await loop.sock_recv(s, 65536)
            v = memoryview(pdata)
            length = 1
            while length and len(v):
                length, typ = struct.unpack('IH', v[:struct.calcsize('IH')])
                if typ == NLMSG_DONE:
                    pending = False
                    break
                if typ == RTM_NEWLINK:
                    ent = v[nlhdrsize:]
                    _, iftyp, ifidx, iflags, _ = struct.unpack('BHiII', ent[:ilen])
                    if iftyp == ARPHRD_ETHER and not iflags & IFF_MASTER:
                        ifname = None
                        rta = ent[ilen:]
                        rtalen = 1
                        while rtalen and len(rta):
                            rtalen, rtatyp = struct.unpack('HH', rta[:4])
                            if rtatyp == IFLA_IFNAME:
                                ifname = rta[4:rtalen].tobytes()
                                ifname = ifname.strip(b'\x00')
                                if not isinstance(ifname, str):
                                    ifname = ifname.decode('utf8')
                                break
                            rta = rta[msg_align(rtalen):]
                        if ifname:
                            idxtoname[ifidx] = ifname
                v = v[msg_align(length):]
        await loop.sock_sendall(s, nlhdr + fdbmg)
        pending = True
        while pending:
            pdata = await loop.sock_recv(s, 65536)
            v = memoryview(pdata)
            length = 1
            while length and len(v):
                length, typ = struct.unpack('IH', v[:struct.calcsize('IH')])
                if typ == NLMSG_DONE:
                    pending = False
                    break
                if typ == RTM_NEWNEIGH:
                    ent = v[nlhdrsize:]
                    fam, idx, state, flags, typ = struct.unpack('BiHBB', ent[:ndmsgsize])
                    if typ == RTN_UNSPEC and not state & NUD_PERMAMENT:
                        curraddr = None
                        currport = idxtoname.get(idx, None)
                        rta = ent[ndmsgsize:]
                        rtalen = 1
                        while rtalen and len(rta):
                            rtalen, rtatyp = struct.unpack('HH', rta[:4])
                            if rtatyp == 2:
                                curraddr = rta[4:rtalen].tobytes()
                                if len(curraddr) == 20:
                                    curraddr = curraddr[12:]
                            elif rtatyp == 1:
                                currip = rta[4:rtalen].tobytes()
                            rta = rta[msg_align(rtalen):]
                        if curraddr and currport:
                            fdb[curraddr] = currport
                v = v[msg_align(length):]
    finally:
        s.close()
    for ipaddr in neightable:
        port = fdb.get(neightable[ipaddr], None)
        if port:
            iptoport[ipaddr] = port

async def get_port_by_mac(macaddr):
    macaddr = bytes(bytearray([int(x, 16) for x in macaddr.split(':')]))
    updated = False
    if neighvintage < os.times()[4] - 30:
        await refresh_neighbor_table()
        updated = True
    port = fdb.get(macaddr, None)
    if not port and not updated:
        await refresh_neighbor_table()
        port = fdb.get(macaddr, None)
    return port

async def get_port(ipaddr):
    ipaddr = ipaddr.split('%', 1)[0]
    port = None
    if ':' in ipaddr:
        ipaddr = socket.inet_pton(socket.AF_INET6, ipaddr)
    elif '.' in ipaddr:
        ipaddr = socket.inet_pton(socket.AF_INET, ipaddr)
    updated = False
    if neighvintage < os.times()[4] - 30:
        await refresh_neighbor_table()
        updated = True
    port = iptoport.get(ipaddr, None)
    if not port and not updated:
        await refresh_neighbor_table()
        port = iptoport.get(ipaddr, None)
    return port

if __name__ == '__main__':
    import sys
    print(repr(asyncio.run(get_port(sys.argv[1]))))