import asyncio
import glob
import socket
import struct

def _sock_recvfrom(loop, fut, rfd, sock, n):
    if rfd is not None:
        loop.remove_reader(rfd)
    if fut.cancelled():
        return
    try:
        retdata = sock.recvfrom(n)
    except (BlockingIOError, InterruptedError):
        fd = sock.fileno()
        loop.add_reader(fd, _sock_recvfrom, loop, fut, fd, sock, n)
    except Exception as exc:
        fut.set_exception(exc)
    else:
        fut.set_result(retdata)

def sock_recvmsg(loop, sock, n):
    if sock.gettimeout() != 0:
        raise ValueError("Blocking!")
    fut = loop.create_future()
    _sock_recvfrom(loop, fut, None, sock, n)
    return fut

def list_interface_indexes():
    for idxfile in glob.glob('/sys/class/net/*/ifindex'):
        with open(idxfile, 'r') as idxin:
            yield int(idxin.read())

async def watch_ssdpv6(handler):
    s = socket.socket(socket.AF_INET6, socket.SOCK_DGRAM)
    s.settimeout(0)
    s.setsockopt(socket.IPPROTO_IPV6, socket.IPV6_V6ONLY, 1)
    for idx in list_interface_indexes():
        v6grp = socket.inet_pton(socket.AF_INET6, 'ff02::c') + struct.pack('=I', idx)
        s.setsockopt(socket.IPPROTO_IPV6, socket.IPV6_JOIN_GROUP, v6grp)
    return await watch_ssdp(s, handler)


async def watch_ssdp(s, handler):
    s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
    s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEPORT, 1)
    s.bind(('', 1900))
    loop = asyncio.get_running_loop()
    running = True
    while running:
        rsp, peer = await sock_recvmsg(loop, s, 2048)
        for line in rsp.split(b'\r\n'):
            if line.startswith(b'NTS: '):
                if line != b'NTS: ssdp:alive':
                    break
            elif line.startswith(b'NT: '):
                if line != b'NT: urn:dmtf-org:service:redfish-rest:1:1':
                    break
            elif line.startswith(b'LOCATION: '):
                if b'/DeviceDescription.json' not in line:
                    break
        else:
            await handler((rsp, peer))

if __name__ == '__main__':
    import pprint
    asyncio.run(watch_ssdpv6(pprint.pprint))

