function b64tobuf(inb: string): Uint8Array {
    inb = inb.replace(/_/g, '/').replace(/-/g, '+')
    return Uint8Array.from(atob(inb), c => c.charCodeAt(0))
}

function buftob64(ins: ArrayBuffer): string {
    return btoa(String.fromCharCode(...new Uint8Array(ins))).replace(/\//g, '_').replace(/\+/g, '-')
}

export class AuthInfo {
    secrets: Map<string,string>;
    password: string;
    username: string;
    constructor(username: string) {
        this.password = '';
        this.username = username;
        this.secrets = new Map();
    }

    add_secret(name: string, value: string): void {
        this.secrets.set(name, value)
    }

    set_password(password: string): void {
        this.password = password;
    }

    get_auth_header(): string {
        if (this.secrets.size > 0) {
            return "MultiBasic " + btoa(this.username + ":" + 
                    JSON.stringify(this.secrets));
        } else if (this.password.length > 0) {
            return "Basic " + btoa(this.username + ":" + 
                this.password);
        }
        return '';
    }
}

export class Session {
    authtoken: string
    sessionid: string
    headers: HeadersInit


    constructor() {
        this.authtoken = ''
        this.sessionid = ''
        this.headers = {
            'SuppressAuthHeader': "true",
            'Accept': 'application/json',
            'ConfluentAuthToken': this.authtoken,
        }
    }

    async fetch(url: string, body: object | undefined = undefined, method: string = '') {
        let payload: string | undefined = undefined
        if (body) {
            this.headers['Content-Type'] = 'application/json'
            if (!method) { method = 'POST' }
            payload = JSON.stringify(body)
        }
        if (!method) { method = 'GET' }
        let result = await fetch(url, {method: method, headers: this.headers, body: payload})
        if (result.ok) {
            return await result.json()
        }
        throw "Error fetching " + url
    }

    async login(authinfo: AuthInfo) {
        let username = authinfo.username
        let auth_header = authinfo.get_auth_header()
        if (auth_header.length > 0) {
            this.headers['Authorization'] = auth_header
            let result = await this.fetch('/confluent-api/sessions/current/info')
            this.sessionid = result['sessionid']
            this.authtoken = result['authtoken']
            delete this.headers['Authorization']
            this.headers['ConfluentAuthToken'] = this.authtoken
        } else {
            // Check if a webauthn is possible
            try {
                let creds = await this.fetch(
                    "/confluent-api/sessions/current/webauthn/registered_credentials/" + username,
                    {}, 'POST');
                creds.challenge = b64tobuf(creds.challenge)
                for (let i of creds.allowCredentials) {
                    i.id = b64tobuf(i.id)
                }
                creds.userVerification = "discouraged"
                let tmpcred = await navigator.credentials.get({publicKey: creds})
                if (tmpcred == undefined) { throw "No credentials registereed" }
                let navcred: PublicKeyCredential = tmpcred as PublicKeyCredential
                let credreq = {}
                credreq['raw_id'] = buftob64(navcred.rawId)
                let authrsp: AuthenticatorAssertionResponse = navcred.response as AuthenticatorAssertionResponse
                credreq['authenticator_data'] = buftob64(authrsp.authenticatorData)
                credreq['client_data_json'] = buftob64(authrsp.clientDataJSON)
                credreq['signature'] = buftob64(authrsp.signature)
                if (authrsp.userHandle != null) {
                    credreq['user_handle'] = buftob64(authrsp.userHandle)
                }
                let authresult = await this.fetch('/confluent-api/sessions/current/webauthn/validate/' + name, credreq, 'POST')
            } catch (e) {
                console.log(e)
            }
        }
    }
}